"""Simple utils file to be shared across backend"""
import importlib
import os
import sys
from queue import Queue
from types import ModuleType
from typing import Callable

# TODO (rossdan): use common sentinel value in HF extension for end of
# stream instead of implicitly relying on None
STOP_STREAMING_SIGNAL = object()
EXCLUDE_OPTIONS = {
    "prompt_index": True,
    "file_path": True,
    "callback_manager": True,
}
LASTMILE_BASE_URI = "https://lastmileai.dev/"


def show_debug():
    """Flip this to True to see debug messages"""
    return False


def get_validated_path(raw_path: str) -> str:
    resolved = convert_to_absolute_path(raw_path)
    if not os.path.isfile(resolved):
        raise Exception(f"File does not exist: {resolved}")
    return resolved


def import_module_from_path(path_to_module: str) -> ModuleType:
    resolved_path = convert_to_absolute_path(path_to_module)
    module_name = os.path.basename(resolved_path).replace(".py", "")

    spec = importlib.util.spec_from_file_location(module_name, resolved_path)
    if spec is None:
        raise Exception(f"Could not import module from path: {resolved_path}")
    elif spec.loader is None:
        raise Exception(
            f"Could not import module from path: {resolved_path} (no loader)"
        )
    else:
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)
        return module


def load_register_fn_from_user_module(
    user_module: ModuleType,
) -> Callable[[], None]:
    if not hasattr(user_module, "register_model_parsers"):
        raise Exception(
            f"User module {user_module} does not have a register_model_parsers function."
        )
    register_fn = getattr(user_module, "register_model_parsers")
    if not callable(register_fn):
        raise Exception(
            f"User module {user_module} does not have a register_model_parsers function"
        )
    else:
        return register_fn


def load_user_parser_module(path_to_module: str) -> None:
    print(f"Importing parsers module from {path_to_module}")
    res_user_module = import_module_from_path(path_to_module)
    register_fn = load_register_fn_from_user_module(res_user_module)
    register_fn()


def convert_to_absolute_path(path: str) -> str:
    return os.path.abspath(os.path.expanduser(path))


class QueueIterator:
    """In order to support text streaming, we need to store
    the output in a queue and iterate over those values. A lot of this was
    inspired by HuggingFace's TextIteratorStreamer object:

    I know I can just use a queue directly in the callsite with
    `iter(queue.get, None)`, but having a class makes it easier to manage
    and abstracts it a bit more.
    """

    def __init__(self):
        self.q = Queue()
        self.stop_signal = STOP_STREAMING_SIGNAL
        self.timeout = None

    def __iter__(self):
        return self

    def __next__(self):
        value = self.q.get(block=True, timeout=self.timeout)
        if value == self.stop_signal:
            raise StopIteration()
        return value

    def put(self, text: str, stream_end: bool = False):
        self.q.put(text, timeout=self.timeout)
        if stream_end:
            self.q.put(self.stop_signal, timeout=self.timeout)

    def isEmpty(self) -> bool:
        return self.q.empty()



import time

from aiconfig import AIConfigRuntime
from pydantic import BaseModel


class SessionData(BaseModel):
    """
    Represents the data stored per session_id, including:
        - AIConfigRuntime: a copied version of the original AIConfigRuntime
        - update_time: the last time this config was updated
            - if it's been after certain amount of time, we will delete it
    """

    config: AIConfigRuntime
    update_time: float = time.time()

    def __lt__(self, other):
        return self.update_time < other.update_time



import json

from .aiconfig_manager import AIConfigManager
from .events import EventHandler
from .gradio_notebook_component import GradioNotebookComponent
from .utils import show_debug


class GradioNotebook:
    """
    A wrapper that constructs a GradioNotebookComponent. We need to do this instead of
    using the GradioNotebookComponent directly because if we try to override the __init__
    method of GradioNotebookComponent, it doesn't properly pass the value into the Component
    """

    config_manager: AIConfigManager

    def __init__(self, config_path: str = "", parsers_path: str = "", **kwargs):
        """
        Args:
            config_path: (str, optional): The filepath to load the AIConfig from.
                If the filepath doesn't exist, a new AIConfig will be created
            parsers_path: (str, optional): The filepath to a module which registers
                additional ModelParsers to the ModelParserRegistry. If the module
                doesn't exist, the default ModelParsers (HF remote inference) will be
                used
        """
        # Create AIConfigManager
        self.config_manager = AIConfigManager(config_path, parsers_path)
        if show_debug():
            print(f"{self.config_manager=}")
            print(f"{self.config_manager.config=}")

        # Display components
        self.component = GradioNotebookComponent(
            value=json.dumps(
                {
                    "aiconfig": self.config_manager.get_config_json("original"),
                    "model_ids": self.config_manager.get_models(),
                }
            ),
            **kwargs,
        )

        # Hook up events
        self._register_events(self.component)

    def _register_events(self, component: GradioNotebookComponent) -> None:
        """
        Register all the events that we need to handle
        Please see the diagram here for visual representation on the steps below:
            https://drive.google.com/file/d/1MgQlh9rL319QBABHuooRnSRymTWoKiEF/view?usp=sharing
        1. Every Component has an EVENTS field which holds the names of events that can
            be triggered on that component (ex: GradioNotebookComponent has
            `add_prompt` event in the EVENTS field)
        2. When we run the app (`gradio cc dev`), this will automagically create an
            .pyi file for each custom component. `.pyi` means Python-interface, and this
            is what is actually being read/run/whatever when using the app. If you open
            this file, you will see that a new function is created for each event
            added in the events field. Ex: `gradio_notebook_component.pyi` contains the
            function `add_prompt()`
        3. This function `add_prompt()` takes in 3 main arguments (many others kwargs,
            but ignore those for now): (<some_function>, <some_inputs>, <some_outputs>)
            Ex: in this function below we setup this piece of code:
                `component.add_prompt(event_handler.add_prompt_impl, [], [component])`
            This means that when the event `add_prompt` is triggered, we will call the
                `event_handler.add_prompt_impl` function by passing the inputs (empty
                in our case) to this function, and return the result of
                `event_handler.add_prompt_impl` to the outputs (components)
        4. The frontend (Index.svelte) dispatches an event corresponding to the name
            of an event in the EVENTS field, and also sends in a json payload which
            contains data that matches the EventData class for that event
        5. The backend defines the EventData class for each event, which specifies
            the schema of the json payload that the frontend sends
            (ex: AddPromptEventData has fields for `prompt` and `existing_ai_config`)
        6. The <some_function> takes in this payload through an EventData object and
            does whatever it wants with it, returning the output which gets sent to
            the output component in part 3
        7. The backend uses the auto-generated event function from part 3 to register
            event listeners using the implementation from part 6
        """
        event_handler = EventHandler(config_manager=self.config_manager)
        component.add_prompt(
            event_handler.add_prompt_impl, [], [component], show_progress=False
        )
        component.cancel_run(
            event_handler.cancel_run_impl, [], [component], show_progress=False
        )
        component.clear_outputs(
            event_handler.clear_outputs_impl, [], [component], show_progress=False
        )
        component.delete_prompt(
            event_handler.delete_prompt_impl, [], [component], show_progress=False
        )
        component.get_aiconfig(
            event_handler.get_aiconfig_impl, [], [component], show_progress=False
        )
        component.remove_session_id(
            event_handler.remove_session_id_impl, [], [component], show_progress=False
        )
        component.run_prompt(
            event_handler.run_prompt_impl, [], [component], show_progress=False
        )
        component.set_config_description(
            event_handler.set_config_description_impl,
            [],
            [component],
            show_progress=False,
        )
        component.set_config_name(
            event_handler.set_config_name_impl, [], [component], show_progress=False
        )
        component.set_parameters(
            event_handler.set_parameters_impl, [], [component], show_progress=False
        )
        component.share_config(
            event_handler.share_config_impl, [], [component], show_progress=False
        )
        component.update_model(
            event_handler.update_model_impl, [], [component], show_progress=False
        )
        component.update_prompt(
            event_handler.update_prompt_impl, [], [component], show_progress=False
        )



import gradio as gr
from gradio_notebook import GradioNotebook

AICONFIG_FILE_PATH = "./example.aiconfig.json"  # Can also be empty or None!
with gr.Blocks() as demo:
    GradioNotebook(config_path=AICONFIG_FILE_PATH)

demo.queue().launch()



from gradio.components.base import Component
from gradio.data_classes import GradioModel
from gradio.events import EventListener

from .utils import show_debug


class GradioAIConfig(GradioModel):
    # TODO(rossdanlm): Explicitly define AIConfigRuntime
    # https://github.com/lastmile-ai/gradio-workbook/issues/62
    pass


class GradioNotebookComponent(Component):
    data_model = GradioAIConfig

    add_prompt_event_listener = EventListener(
        "add_prompt",
        doc="Triggered when user clicks the Add Prompt button",
    )
    cancel_run_event_listener = EventListener(
        "cancel_run",
        doc="Triggered when user clicks the cancel button while a prompt is running",
    )
    clear_outputs_event_listener = EventListener(
        "clear_outputs",
        doc="Triggered when user clicks the Clear Outputs button",
    )
    delete_prompt_event_listener = EventListener(
        "delete_prompt",
        doc="Triggered when user clicks the Delete Prompt button",
    )
    get_aiconfig_event_listener = EventListener(
        "get_aiconfig",
        doc="Triggered when user needs a non-stale version of config from server to perform an action directly on client (ex: download)",
    )
    remove_session_id_event_listener = EventListener(
        "remove_session_id",
        doc="Triggered when window is closing and session is about to be deleted",
    )
    share_config_event_listener = EventListener(
        "share_config",
        doc="Triggered when user clicks the Share button",
    )
    run_prompt_event_listener = EventListener(
        "run_prompt",
        doc="Triggered when user clicks the Run Prompt button",
    )
    set_config_description_event_listener = EventListener(
        "set_config_description",
        doc="Triggered when user sets the description for the AIConfig",
    )
    set_config_name_event_listener = EventListener(
        "set_config_name",
        doc="Triggered when user sets the name for the AIConfig",
    )
    set_parameters_event_listener = EventListener(
        "set_parameters",
        doc="Triggered when user sets global or prompt parameters. If done with a prompt, \
            it will update the param for that prompt. Otherwise, it will update \
            the overall parameters for the AIConfig",
    )
    update_model_event_listener = EventListener(
        "update_model",
        doc="Triggered when user clicks the Update Model button. If done with a prompt, \
            it will update the model settings for that prompt. Otherwise, it will update \
            the overall model settings for the AIConfig",
    )
    update_prompt_event_listener = EventListener(
        "update_prompt",
        doc="Triggered when user updates the prompt name or settings",
    )
    EVENTS = [
        add_prompt_event_listener,
        cancel_run_event_listener,
        clear_outputs_event_listener,
        delete_prompt_event_listener,
        get_aiconfig_event_listener,
        remove_session_id_event_listener,
        run_prompt_event_listener,
        share_config_event_listener,
        set_config_description_event_listener,
        set_config_name_event_listener,
        set_parameters_event_listener,
        update_model_event_listener,
        update_prompt_event_listener,
    ]

    def preprocess(self, payload: GradioAIConfig | None):
        if show_debug():
            print("\nin preprocess")
            print(f"{payload=}")
        if payload is None:
            return payload
        return payload

    def postprocess(self, value: str) -> str:
        """
        Return an output in the form of a json-formatted string for
        client to process
        """
        if show_debug():
            print("\nin postprocess")
            print(f"{value=}")
        return value

    # TODO (rossdanlm): Update this with a valud AIConfigRuntime object
    # https://github.com/lastmile-ai/gradio-workbook/issues/62
    def example_inputs(self):
        if show_debug():
            print("\nin example_inputs")
        return {"foo": "bar"}

    def api_info(self):
        return {"type": {}, "description": "AIConfigRuntime"}
